"""Resource class that forms the base of all DID Document components."""
from abc import ABC, abstractmethod
import json
from typing import Any, Dict, Type, TypeVar

from inflection import camelize
from pydantic import BaseModel, Extra, parse_obj_as
from typing_extensions import Literal
import typing_extensions

from .validation import wrap_validation_error


ResourceType = TypeVar("ResourceType", bound="Resource")


if hasattr(typing_extensions, "get_args"):  # pragma: no cover
    from typing_extensions import get_args, get_origin

    def get_literal_values(literal):
        """Return the args of a literal."""
        return get_args(literal)

    def is_literal(type_):
        """Return if type is literal."""
        return get_origin(type_) is Literal

else:  # pragma: no cover
    # Python 3.6 and Literals behave differently
    from typing_extensions import _Literal

    def get_literal_values(literal):
        """Return the args of a literal."""
        return literal.__values__

    def is_literal(type_):
        """Return if type is literal."""
        return isinstance(type_, _Literal)


class Resource(BaseModel):
    """Base class for DID Document components."""

    class Config:
        """Configuration for Resources."""

        underscore_attrs_are_private = True
        extra = Extra.allow
        allow_population_by_field_name = True
        allow_mutation = False

        @classmethod
        def alias_generator(cls, string: str) -> str:
            """Transform snake_case to camelCase."""
            return camelize(string, uppercase_first_letter=False)

    def serialize(self):
        """Return serialized representation of Resource."""
        return self.dict(exclude_none=True, by_alias=True)

    @classmethod
    def deserialize(cls: Type[ResourceType], value: dict) -> ResourceType:
        """Deserialize into VerificationMethod."""
        with wrap_validation_error(
            ValueError,
            message="Failed to deserialize {}".format(cls.__name__),
        ):
            return parse_obj_as(cls, value)

    @classmethod
    def from_json(cls, value: str):
        """Deserialize Resource from JSON."""
        loaded: dict = json.loads(value)
        return cls.deserialize(loaded)

    def to_json(self):
        """Serialize Resource to JSON."""
        return self.json(exclude_none=True, by_alias=True)

    @classmethod
    def _fill_in_required_literals(cls, **kwargs) -> Dict[str, Any]:
        """Return dictionary of field name to value from literals."""
        for field in cls.__fields__.values():
            if (
                field.required
                and is_literal(field.type_)
                and (field.name not in kwargs or kwargs[field.name] is None)
            ):
                kwargs[field.name] = get_literal_values(field.type_)[0]
        return kwargs

    @classmethod
    def _overwrite_none_with_defaults(cls, **kwargs) -> Dict[str, Any]:
        """Overwrite none values in kwargs with defaults for corresponding field."""
        for field in cls.__fields__.values():
            if field.name in kwargs and kwargs[field.name] is None:
                kwargs[field.name] = field.get_default()
        return kwargs

    @classmethod
    def make(cls: Type[ResourceType], **kwargs) -> ResourceType:
        """Create instance of class, filling in literals."""
        kwargs = cls._fill_in_required_literals(**kwargs)
        kwargs = cls._overwrite_none_with_defaults(**kwargs)
        return cls(**kwargs)


class IndexedResource(Resource, ABC):
    """Resource with index for supporting dereferencing nested objects."""

    _index: dict = {}

    def __init__(self, **data):
        super().__init__(**data)
        self._index_resources()

    @abstractmethod
    def _index_resources(self):
        """Index nested resources."""

    @abstractmethod
    def dereference(self, reference: str) -> Resource:
        """Dereference a nested object."""

    def dereference_as(self, typ: Type[ResourceType], reference: str) -> ResourceType:
        """Dereference a resource to a specific type."""
        resource = self.dereference(reference)
        try:
            return parse_obj_as(typ, resource.dict())
        except ValueError as error:
            raise ValueError(
                "Dereferenced resource {} could not be parsed as {}".format(
                    resource, typ
                )
            ) from error

    @classmethod
    def construct(cls, **data):
        """Construct and index."""
        resource = super(Resource, cls).construct(**data)
        resource._index_resources()
        return resource



"""Validation tools and helpers."""

from contextlib import contextmanager
from typing import Set, Type

from pydantic import ValidationError, root_validator


@contextmanager
def wrap_validation_error(error_to_raise: Type[Exception], message: str = None):
    """Wrap validation erros with more friendly errors."""
    try:
        yield
    except ValidationError as error:
        raise error_to_raise(
            ":\n".join([message, str(error)]) if message else str(error)
        ) from error


def required_group(props: Set[str]):
    """Require at least one of the properties to be present."""

    def _require_group(_model, values: dict):
        defined_props = props & {
            key for key, value in values.items() if value is not None
        }
        if len(defined_props) < 1:
            raise ValueError(
                "At least one of {} was required; none found".format(props)
            )
        return values

    return root_validator(allow_reuse=True)(_require_group)



"""DID Doc Service."""

from typing import List, Optional, Union

from pydantic import Extra, AnyUrl
from typing_extensions import Literal

from .did_url import DIDUrl
from .resource import Resource


class ServiceEndpoint(Resource):
    """List of Service Endpoints"""

    uri: Union[DIDUrl, AnyUrl]


class Service(Resource):
    """Representation of DID Document Services."""

    id: DIDUrl
    type: str
    service_endpoint: Union[DIDUrl, AnyUrl, Literal[""], List[ServiceEndpoint]]


class DIDCommService(Service):
    """DID Communication Service."""

    class Config:
        """DIDComm Service Config."""

        extra = Extra.forbid

    type: Literal["IndyAgent", "did-communication"] = "did-communication"
    recipient_keys: List[DIDUrl]
    routing_keys: List[DIDUrl] = []
    accept: Optional[List[str]] = None
    priority: int = 0


class UnknownService(Service):
    """Unknown Service."""
